/*
 * File: buffer.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 * MACOS port by Aria Stewart <aredridel@nbtsc.org>
 */

/***

Buffer management functions.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"buffer.h"
#include	"bind.h"
#include	"cursor.h"
#include	"display.h"
#include	"file.h"
#include	"line.h"
#include	"messages.h"
#include        "terminal.h"
#include	"utils.h"
#include	"window.h"



/*++

Global helper to move a loaded BUFFER bp to top of buffer list.

++*/
int make_bp_top(bp)
BUFFER *bp;

{
    WNDW *wp;
    SCRN *scrp;                 /* screen to fix pointers in */
    int cmark;                  /* current mark */

    /* allow exit-buffer hook macro */
    execute_key(&exbhook, FALSE, 1);

    if (--curbp->b_nwnd == 0) { /* Last use.            */
        curbp->b_dotp = curwp->w_dotp;
        curbp->b_doto = curwp->w_doto;
        for (cmark = 0; cmark < NMARKS; cmark++) {
            curbp->b_markp[cmark] = curwp->w_markp[cmark];
            curbp->b_marko[cmark] = curwp->w_marko[cmark];
        }
        curbp->b_fcol = curwp->w_fcol;
    }
    curbp = bp;                 /* Switch */
    if (curbp->b_active != TRUE) {      /* buffer not active yet */
        /* read it in and activate it */
        readin(curbp->b_name, ((curbp->b_mode & MDVIEW) == 0));
        curbp->b_dotp = lforw(curbp->b_linep);
        curbp->b_doto = 0;
        curbp->b_active = TRUE;
    }
    curwp->w_bufp = bp;
    curwp->w_linep = bp->b_linep;       /* For macros, ignored. */
    curwp->w_flag |= WFMODE | WFFORCE | WFHARD;
    if (bp->b_nwnd++ == 0) {    /* First use */
        curwp->w_dotp = bp->b_dotp;
        curwp->w_doto = bp->b_doto;
        for (cmark = 0; cmark < NMARKS; cmark++) {
            curwp->w_markp[cmark] = bp->b_markp[cmark];
            curwp->w_marko[cmark] = bp->b_marko[cmark];
        }
        curwp->w_fcol = bp->b_fcol;
    } else {                    /* in all screens ... */
        scrp = first_screen;
        while (scrp) {
            wp = scrp->s_first_window;
            while (wp != NULL) {
                if (wp != curwp && wp->w_bufp == bp) {
                    curwp->w_dotp = wp->w_dotp;
                    curwp->w_doto = wp->w_doto;
                    for (cmark = 0; cmark < NMARKS; cmark++) {
                        curwp->w_markp[cmark] = wp->w_markp[cmark];
                        curwp->w_marko[cmark] = wp->w_marko[cmark];
                    }
                    curwp->w_fcol = wp->w_fcol;
                    break;
                }
                /* next window */
                wp = wp->w_wndp;
            }
            /* next screen! */
            scrp = scrp->s_next_screen;
        }
    }

    /* allow enter-buffer hook macro */
    execute_key(&bufhook, FALSE, 1);

    return (TRUE);
}



/*++

Global helper for getting the next non-invisible
buffer after the current one.

+**/
BUFFER *get_default_buffer()
{
    BUFFER *bp;                 /* default buffer */

    /* Find the next buffer, which will be the default */
    bp = curbp->b_bufp;

    /* cycle through the buffers to find an eligible one */
    while (bp == NULL || bp->b_flag & BFINVS) {
        if (bp == NULL)
            bp = bheadp;
        else
            bp = bp->b_bufp;

        /* don't get caught in an infinite loop! */
        if (bp == curbp) {
            bp = NULL;
            break;
        }
    }
    return (bp);
}



/*++

Removes all text from buffer bp. If the buffer is marked as changed then
we ask first. The window chain is nearly always wrong when this is called;
caller must arrange for any updates that are required.

++*/
int clear_buffer(bp)
BUFFER *bp;

{
    LINE *lp;
    int s;
    int cmark;                  /* current mark */

    if ((bp->b_flag & BFINVS) == 0      /* Invisible buffer               */
        && (bp->b_flag & BFCHG) != 0    /* -OR- changed but the            */
        && (s = mlyesno(TEXT32)) != TRUE)       /* "Discard changes" not accepted */
        return (s);
    bp->b_flag &= ~BFCHG;       /* mark as unchanged  */
    while ((lp = lforw(bp->b_linep)) != bp->b_linep)
        free_line(lp);          /* Free lines of text */
    bp->b_dotp = bp->b_linep;   /* Fix "."            */
    bp->b_doto = 0;
    for (cmark = 0; cmark < NMARKS; cmark++) {
        bp->b_markp[cmark] = NULL;      /* Invalidate marks   */
        bp->b_marko[cmark] = 0;
    }
    bp->b_fcol = 0;
    return (TRUE);
}



/*++

Release BUFFER bp from the buffer list.  No error-handlng here.
Send releasable buffer or segfault.

++*/
int release_buffer(bp)
BUFFER *bp;

{
    BUFFER *bp1;
    BUFFER *bp2;

    free((char *) bp->b_linep); /* Release header line. */
    bp1 = NULL;                 /* Find the header.     */
    bp2 = bheadp;
    while (bp2 != bp) {
        bp1 = bp2;
        bp2 = bp2->b_bufp;
    }
    bp2 = bp2->b_bufp;          /* Next one in chain.   */
    if (bp1 == NULL)            /* Unlink it.           */
        bheadp = bp2;
    else
        bp1->b_bufp = bp2;
    free((char *) bp);          /* Release buffer block */
    return (TRUE);
}



/*++

Switch to the next (default) buffer in the buffer list.

++*/
int next_buffer(f, n)
int f, n;
{
    BUFFER *bp;
    int status;

    /* make sure the arg is legit */
    if (f == FALSE)
        n = 1;
    if (n < 1)
        return (FALSE);

    /* cycle thru buffers until n runs out */
    while (n-- > 0) {
        bp = get_default_buffer();
        /* skip scratch buffer */
        if (strcmp(bp->b_name, mainbuf) == 0) {
            make_bp_top(bp);
            bp = get_default_buffer();
        }
        if (bp == NULL)
            return (FALSE);
        status = make_bp_top(bp);
        if (status != TRUE)
            return (status);
    }
    return (status);
}



/*++

Dispose of a buffer, by name.  Clear the buffer.
Then free the header line and the buffer header.

++*/
int delete_buffer(f, n)
int f, n;

{
    BUFFER *bp;                 /* ptr to buffer to dump */
    BUFFER *bnext;              /* buffer below curbp */
    char name[LENPATH];
    int ok = FALSE;             /* okay to delete changed buffer */

    f = 0;                      /* splint */
    bp = curbp;

    /* do not kill an executing buffer */
    if (bp->b_exec != 0) {
        mlwrite(TEXT226);       /* "%%Can not delete an executing buffer" */
        return (FALSE);
    }

    single_window(0, 0);
    strcpy(name, curbp->b_name);

    /* so we don't run off the end */
    (void) exist_buffer(mainbuf, TRUE, 0);

    /* get the buffer name to kill */
    bp = prompt4buffer(TEXT26, bp ? bp->b_name : mainbuf, TRUE);        /*"Kill buffer" */
    if (bp == NULL)
        return (ABORT);

    /* prevents crashing on last buffer */
    if (strcmp(bp->b_name, mainbuf) == 0) {
        mlwrite(TEXT4);         /* Cannot delete scratch buffer */
        return (FALSE);
    }

    /* Blow bp text away? */
    if ((ok = clear_buffer(bp)) != TRUE)
        return (ok);

    /* if we're curbp, submerge */
    if (strcmp(name, bp->b_name) == 0) {
        bnext = get_default_buffer();
        make_bp_top(bnext);
    }

    n = delbufbyname(bp->b_name);       /* Here we delete the buffer */
    if (n == ABORT)
        return (n);

    /* scratch curbp if only bp */
    if (strcmp(curbp->b_name, mainbuf) == 0) {
        bnext = get_default_buffer();
        if (bnext != NULL)
            make_bp_top(bnext);
    }
    return (TRUE);
}



/*++

Non-interactive delete_buffer.  Send name of releasable
buffer or segfault.

++*/
int delbufbyname(name)
char *name;
{
    BUFFER *bp;
    bp = exist_buffer(name, FALSE, 0);
    if (bp == NULL)
        return (ABORT);

    return (release_buffer(bp));
}

#if MACOS
/* nop */
#else
extern int asprintf(char **, char *, ...);
#endif                          /* ! MACOS */


/*++

Attach a buffer to a single window. Lists buffers and allows user to
select or abort.  Out-of-range selections fall back to current buffer.
This routine rebuilds the text in the hidden buffer that holds the
buffer list.

++*/
int list_buffers()
{

    BUFFER *bp;                 /* for list handling */
    BUFFER *lastbp;             /* for returning to curbp */
    char *cp1;                  /* two ptrs for line-building */
    char *cp2;
    char *count;
    char choice[NLINE];         /* buffer you want to change to */
    char line[128];             /* displayed in list */
    int i, j;                   /* choice loop */
    int s;                      /* mlreply status */
    int ok = FALSE;             /* buffer is chosen */
    int numbufs = 1;            /* and incremented in display */
    int nbytes;                 /* # of bytes in current buffer */
    LINE *lp;                   /* lines we add for each buffer */

    lastbp = curbp;

    blistp->b_flag &= ~BFCHG;
    clear_buffer(blistp);       /* Blow old text away   */
    if (addline(blistp, TEXT30) == FALSE        /*"ACTN Modes Size Buffer File" */
        || addline(blistp,
                   "------------------------------------------ ") == FALSE)
        return (FALSE);

    /* build line to report global mode settings */
    cp1 = &line[0];
    for (i = 0; i < 8; i++)
        *cp1++ = ' ';

    /* output the mode codes */
    for (i = 0; i < NUMMODES; i++)
        if (gmode & (1 << i))
            *cp1++ = modecode[i];
        else
            *cp1++ = '.';
    *cp1++ = ' ';
    strcpy(cp1, TEXT31);        /* "Global Modes" */
    if (addline(blistp, line) == FALSE)
        return (FALSE);

    bp = bheadp;                /* For all buffers      */
    while (bp != NULL) {        /* output the list of buffers */
        /* skip invisible buffers */
        if ((!debug) && (bp->b_flag & BFINVS) != 0) {
            bp = bp->b_bufp;
            continue;
        }

        cp1 = &line[0];         /* Start at left edge   */
        *cp1++ = '(';
        count = int2str(numbufs++);
        cp2 = &count[0];
        for (i = 0; i < (int) strlen(count); i++)
            *cp1++ = *cp2++;
        *cp1++ = ')';

        /* output status of ACTIVE flag (has the file been read in? */
        if (bp->b_active == TRUE)       /* "@" if activated       */
            *cp1++ = '@';
        else
            *cp1++ = ' ';

        /* output status of changed flag */
        if ((bp->b_flag & BFCHG) != 0)  /* "*" if changed       */
            *cp1++ = '*';
        else
            *cp1++ = ' ';

        /* report if the file is truncated */
        if ((bp->b_flag & BFTRUNC) != 0)
            *cp1++ = '#';
        else
            *cp1++ = ' ';

        /* report if the file is narrowed */
        if ((bp->b_flag & BFNAROW) != 0)
            *cp1++ = '<';
        else
            *cp1++ = ' ';

        *cp1++ = ' ';           /* space */

        /* output the mode codes */
        for (i = 0; i < NUMMODES; i++) {
            if (bp->b_mode & (1 << i))
                *cp1++ = modecode[i];
            else
                *cp1++ = '.';
        }
        *cp1++ = ' ';           /* Gap.                 */
        nbytes = 0L;            /* Count bytes in buf.  */
        lp = lforw(bp->b_linep);
        while (lp != bp->b_linep) {
            nbytes += lused(lp) + 1;
            lp = lforw(lp);
        }
        asprintf(&cp2, "%8lu", nbytes);
        while (*cp2)
            *cp1++ = *cp2++;
        *cp1++ = ' ';           /* Gap.                 */
        cp2 = &bp->b_name[0];   /* Buffer name          */
        while (*cp2)
            *cp1++ = *cp2++;
        *cp1 = 0;               /* Add to the buffer.   */
        if (addline(blistp, line) == FALSE)
            return (FALSE);
        bp = bp->b_bufp;
    }

    blistp->b_mode |= MDVIEW;
    make_bp_top(blistp);
    update_screen(0, 0);

    /* get buffer choice */
    while (!ok) {
        s = mlreply(TEXT27, choice, NLINE);
        if (!s) {
            make_bp_top(lastbp);
            return (FALSE);
        }
        if (strcmp(choice, "d") == 0) {
            page_down();
            update_screen(0, 0);
            continue;
        }
        if (strcmp(choice, "u") == 0) {
            page_up();
            update_screen(0, 0);
            continue;
        }
        i = str2int(choice) - 1;
        if ((i >= 1) || (i <= numbufs))
            ok = TRUE;
        else
            s = FALSE;
    }

    /* run down buffers to choice ... */
    j = 0;
    bp = bheadp;
    while (bp != NULL) {
        if ((!debug) && (bp->b_flag & BFINVS) != 0) {
            bp = bp->b_bufp;
            continue;
        }
        if (j == i) {
            make_bp_top(bp);
            break;
        }
        j++;
        bp = bp->b_bufp;
    }

    /* ...or revert to the one we rode in on */
    if (bp == NULL)
        make_bp_top(lastbp);

    return (TRUE);
}



/*++

Look for a buffer by name. Return a pointer to the BUFFER structure
associated with it.  If buffer not found and "cflag" TRUE, create
it. The "bflag" sets the flags in buffer.

++*/
BUFFER *exist_buffer(bname, cflag, bflag)
char *bname;                    /* name of buffer to find */
int cflag;                      /* create it if not found? */
int bflag;                      /* bit settings for a new buffer */

{
    BUFFER *bp;
    BUFFER *sb;                 /* buffer to insert after */
    LINE *lp;
    int cmark;                  /* current mark */

    bp = bheadp;
    while (bp != NULL) {
        if (strcmp(bname, bp->b_name) == 0)
            return (bp);
        bp = bp->b_bufp;
    }

    /* no such buffer exists, create it */
    if (cflag != FALSE) {

        /* allocate the needed memory */
        bp = (BUFFER *) malloc(sizeof(BUFFER));

        if ((lp = alloc_line(0)) == NULL) {
            free((char *) bp);
            return (NULL);
        }

        /* find the place in the list to insert this buffer */
        if (bheadp == NULL || strcmp(bheadp->b_name, bname) > 0) {
            /* insert at the beginning */
            bp->b_bufp = bheadp;
            bheadp = bp;
        } else {
            sb = bheadp;
            while (sb->b_bufp != NULL) {
                if (strcmp(sb->b_bufp->b_name, bname) > 0)
                    break;
                sb = sb->b_bufp;
            }

            /* and insert it */
            bp->b_bufp = sb->b_bufp;
            sb->b_bufp = bp;
        }

        /* and set up the other buffer fields */
        bp->b_topline = NULL;
        bp->b_botline = NULL;
        bp->b_active = TRUE;
        bp->b_dotp = lp;
        bp->b_doto = 0;
        for (cmark = 0; cmark < NMARKS; cmark++) {
            bp->b_markp[cmark] = NULL;
            bp->b_marko[cmark] = 0;
        }
        bp->b_fcol = 0;
        bp->b_flag = bflag;
        bp->b_mode = gmode;
        bp->b_nwnd = 0;
        bp->b_exec = 0;
        bp->b_linep = lp;
        strcpy(bp->b_name, bname);
        lp->l_fp = lp;
        lp->l_bp = lp;
    }
    return (bp);
}



/*++

User toggle for the change flag on a buffer.
This should be handled internally when we have undo.

++*/
int unmark_buffer(f, n)
int f, n;

{
    f = n = 0;                  /*  splint  */

    /* unmark_buffer the buffer */
    curbp->b_flag &= ~BFCHG;

    /* unmark_buffer all windows as well */
    update_modeline();

    return (TRUE);
}



/*++

If single buffer bp is on top, bring next buffer up
and release bp

++*/
int bpflip_release(bp)
BUFFER *bp;
{
    BUFFER *bp2;

    if (bp == curbp) {
        bp2 = get_default_buffer();
        make_bp_top(bp2);
    }

    if (!release_buffer(bp))
        return (0);

    return (1);
}



/*++

Attach a buffer to a window. The values of dot and mark come from the
buffer if the use count is 0. Otherwise, they come from some other window.

++*/
int select_buffer(f, n)
int f, n;

{
    BUFFER *bp;                 /* temporary buffer pointer */

    f = n = 0;                  /* splint */


    bp = get_default_buffer();

    /* skip scratch buffer */
    if (strcmp(bp->b_name, mainbuf) == 0) {
        make_bp_top(bp);
        bp = get_default_buffer();
    }

    /* get the buffer name to switch to */
    bp = prompt4buffer(TEXT24, bp ? bp->b_name : mainbuf, TRUE);        /* "Use buffer" */
    if (!bp)
        return (ABORT);

    /* switch to bp */
    return (make_bp_top(bp));
}



/*++

Delete all sourced buffers

++*/
int purge_sbuffers()
{
    BUFFER *bp;
    BUFFER *tmpbp;
    char name[LENPATH];

    bp = bheadp;                /* For all buffers      */
    while (bp != NULL) {

        if (bp->b_flag & BFINVS) {      /* macros are invisible buffers */
            strcpy(name, bp->b_name);
            if ((name[strlen(name) - 1] == ']')
                && (strcmp(name, "[Buffers]") != 0)) {
                clear_sbinding(name);
                tmpbp = bp;
                bp = bp->b_bufp;
                release_buffer(tmpbp);
            } else
                bp = bp->b_bufp;
        } else
            bp = bp->b_bufp;
    }
    return TRUE;
}
