/*
 * File: input.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */

/***
	The methods in this file are those used by the message line itself.
	The file messages.c contains the message line functions used by
	other files.

	From the old editor code:

	MicroEMACS processes two distinct forms of
	characters.  One of these is a standard unsigned character
	which is used in the edited text.  The other form, called
	an Extended Character is a 2 byte value which contains
	both an ascii value, and flags for certain prefixes/events.

	Bit	Usage
	---	-----
	0 -> 7	Standard 8 bit ascii character
	8	Control key flag
	9	META prefix flag
	10	^X prefix flag
	11	Function key flag
	12	(Mouse prefix no longer used)
	13	Shifted flag (not needed on alpha shifted characters)
	14	Alterate prefix (ALT key on PCs)

	The machine dependent driver is responsible for returning
	a byte stream from the various input devices with various
	prefixes/events embedded as escape codes.  Zero is used as the
	value indicating an escape sequence is next.  The format of
	an escape sequence is as follows:

	0		Escape indicator
	<prefix byte>	upper byte of extended character
	{<col><row>}	col, row position if the prefix byte
			indicated a menu selection
			in which case these form a 16 bit menu ID
	<event code>	value of event

	A ^<space> sequence (0/1/32) is generated when an actual
	null is being input from the control-space key under many
	unix systems.  These values are then interpreted by mlgetkey()
	to construct the proper extended character sequences to pass
	to the kernel.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<dirent.h>
#include	<sys/stat.h>

#include	"rho.h"
#include	"input.h"
#include	"buffer.h"
#include	"display.h"
#include	"exec.h"
#include	"input.h"
#include	"line.h"
#include	"messages.h"
#include	"terminal.h"
#include	"utils.h"
#include	"window.h"


static DIR *dirptr = NULL;      /* Current directory stream     */
static char rbuf[LENPATH];      /* Return file buffer           */
static char *nameptr;           /* Ptr past end of path in rbuf */


/*++

Get next filename from pattern

++*/
static char *getnfile()
{
    struct dirent *dp;
    struct stat fstat;

    /* ...and call for the next file */
    do {
        dp = readdir(dirptr);
        if (!dp)
            return (NULL);

        /* Check to make sure we skip all weird entries except directories */
        strcpy(nameptr, dp->d_name);

    }
    while (stat(rbuf, &fstat) ||
           ((fstat.st_mode & S_IFMT) & (S_IFREG | S_IFDIR)) == 0);

    /* if this entry is a directory name, say so */
    if ((mode_t) (fstat.st_mode & S_IFMT) == S_IFDIR)
        strcat(rbuf, DIRSEPSTR);

    /* Return the next file name! */
    return (rbuf);
}




/*++

Get first filename from pattern

++*/
static char *getffile(fspec)
char *fspec;                    /* Filename specification       */
{
    int index;
    char path[LENPATH];         /* Path of file to find         */

    /* First parse the file path off the file spec */
    strcpy(path, fspec);
    index = (int) strlen(path) - 1;
    while (index >= 0 && (path[index] != '/'))
        --index;
    path[index + 1] = '\0';


    /* Open the directory pointer */
    if (dirptr) {
        closedir(dirptr);
        dirptr = NULL;
    }

    dirptr = opendir((path[0] == '\0') ? "./" : path);

    if (!dirptr)
        return (NULL);

    strcpy(rbuf, path);
    nameptr = &rbuf[strlen(rbuf)];

    /* ...and call for the first file */
    return (getnfile());
}





/*++	

Make a completion list based on a partial command name.

++*/
static void possible_command(name, cpos)
char *name;                     /* command containing the current name to complete */
int *cpos;                      /* ptr to position of next character to insert */

{
    NBIND *bp;                  /* trial command to complete */
    int curbind;                /* index into the names[] array */
    int name_len;               /* current length of input string */
    BUFFER *listbuf;            /* buffer to put completion list into */

    /* get a buffer for the completion list */
    listbuf = exist_buffer("[Completion list]", TRUE, BFINVS);
    if (listbuf == NULL || clear_buffer(listbuf) == FALSE) {
        abort_command(FALSE, 0);
        TTflush();
        return;
    }

    name_len = *cpos;

    /* first, we start at the first command and scan the list */
    for (curbind = 0; curbind <= numfunc; curbind++) {

        /* is this a match? */
        bp = &names[curbind];
        if (strncmp(name, bp->n_name, (size_t) name_len) == 0)
            addline(listbuf, bp->n_name);
    }

    wpopup(listbuf);
    return;
}




/*++

Attempt a completion on a command name.

++*/
static void complete_command(name, cpos)
char *name;                     /* command containing the current name to complete */
int *cpos;                      /* ptr to position of next character to insert */

{
    NBIND *bp;                  /* trial command to complete */
    int index;                  /* index into strings to compare */
    int curbind;                /* index into the names[] array */
    NBIND *match;               /* last command that matches string */
    int matchflag;              /* did this command name match? */
    int comflag;                /* was there a completion at all? */

    /* start attempting completions, one character at a time */
    comflag = FALSE;
    curbind = 0;
    while (*cpos < NSTRING) {

        /* first, we start at the first command and scan the list */
        match = NULL;
        curbind = 0;
        while (curbind <= numfunc) {

            /* is this a match? */
            bp = &names[curbind];
            matchflag = TRUE;
            for (index = 0; index < *cpos; index++)
                if (name[index] != bp->n_name[index]) {
                    matchflag = FALSE;
                    break;
                }

            /* if it is a match */
            if (matchflag) {

                /* if this is the first match, simply record it */
                if (match == NULL) {
                    match = bp;
                    name[*cpos] = bp->n_name[*cpos];
                } else {
                    /* if there's a difference, stop here */
                    if (name[*cpos] != bp->n_name[*cpos])
                        return;
                }
            }

            /* on to the next command */
            curbind++;
        }

        /* with no match, we are done */
        if (match == NULL) {
            /* beep if we never matched */
            if (comflag == FALSE)
                TTbeep();
            return;
        }

        /* if we have completed all the way... go back */
        if (name[*cpos] == 0) {
            (*cpos)++;
            return;
        }

        /* remember we matched, and complete one character */
        comflag = TRUE;
        TTputc(name[(*cpos)++]);
        TTflush();
    }

    /* don't allow a completion past the end of the max command name length */
    return;
}



/*++

Attempt a completion on a buffer name

++*/
static void complete_buffer(name, cpos)
char *name;                     /* buffer containing the current name to complete */
int *cpos;                      /* ptr to position of next character to insert */

{
    BUFFER *bp;                 /* trial buffer to complete */
    int index;                  /* index into strings to compare */
    BUFFER *match;              /* last buffer that matches string */
    int matchflag;              /* did this buffer name match? */
    int comflag;                /* was there a completion at all? */

    /* start attempting completions, one character at a time */
    comflag = FALSE;
    while (*cpos < LENPATH) {

        /* first, we start at the first buffer and scan the list */
        match = NULL;
        bp = bheadp;
        while (bp) {

            /* is this a match? */
            matchflag = TRUE;
            for (index = 0; index < *cpos; index++)
                if (name[index] != bp->b_name[index]) {
                    matchflag = FALSE;
                    break;
                }

            /* if it is a match */
            if (matchflag) {

                /* if this is the first match, simply record it */
                if (match == NULL) {
                    match = bp;
                    name[*cpos] = bp->b_name[*cpos];
                } else {
                    /* if there's a difference, stop here */
                    if (name[*cpos] != bp->b_name[*cpos])
                        return;
                }
            }

            /* on to the next buffer */
            bp = bp->b_bufp;
        }

        /* with no match, we are done */
        if (match == NULL) {
            /* beep if we never matched */
            if (comflag == FALSE)
                TTbeep();
            return;
        }

        /* if we have completed all the way... go back */
        if (name[*cpos] == 0) {
            (*cpos)++;
            return;
        }

        /* remember we matched, and complete one character */
        comflag = TRUE;
        TTputc(name[(*cpos)++]);
        TTflush();
    }

    /* don't allow a completion past the end of the max buffer name length */
    return;
}




/*++

Attempt a completion on a file name.

++*/
static void complete_filename(name, cpos)
char *name;                     /* file containing the current name to complete */
int *cpos;                      /* ptr to position of next character to insert */

{
    char *fname;                /* trial file to complete */
    int index;                  /* index into strings to compare */

    int matches;                /* number of matches for name */
    char longestmatch[NSTRING]; /* temp buffer for longest match */
    int longestlen;             /* length of longest match (always > *cpos) */

    /* everything (or nothing) matches an empty string */
    if (*cpos == 0)
        return;

    /* first, we start at the first file and scan the list */
    matches = 0;
    name[*cpos] = 0;
    fname = getffile(name);
    while (fname) {

        /* is this a match? */
        if (strncmp(name, fname, (size_t) * cpos) == 0) {

            /* count the number of matches */
            matches++;

            /* if this is the first match, simply record it */
            if (matches == 1) {
                strcpy(longestmatch, fname);
                longestlen = (int) strlen(longestmatch);
            } else {

                /* if there's a difference, stop here */
                if (longestmatch[*cpos] != fname[*cpos])
                    return;

                for (index = (*cpos) + 1; index < longestlen; index++)
                    if (longestmatch[index] != fname[index]) {
                        longestlen = index;
                        longestmatch[longestlen] = 0;
                    }
            }
        }

        /* on to the next file */
        fname = getnfile();
    }

    /* beep if we never matched */
    if (matches == 0) {
        TTbeep();
        return;
    }

    /* the longestmatch array contains the longest match so copy and print it */
    for (; (*cpos < (NSTRING - 1)) && (*cpos < longestlen); (*cpos)++) {
        name[*cpos] = longestmatch[*cpos];
        TTputc(name[*cpos]);
    }

    name[*cpos] = 0;

    /* if only one file matched then increment cpos to signal complete() */
    /* that this was a complete match.  If a directory was matched then */
    /* last character will be the DIRSEPCHAR.  In this case we do NOT */
    /* want to signal a complete match. */
    if ((matches == 1) && (name[(*cpos) - 1] != DIRSEPCHAR))
        (*cpos)++;

    TTflush();

    return;
}





/*++

This is the message line engine.

++*/
char *complete(prompt, defval, type, maxlen)
char *prompt;                   /* prompt to user on command line */
char *defval;                   /* default value to display to user */
int type;                       /* type of what we are completing */
int maxlen;                     /* maximum length of input field */

{
    int c;                      /* current input character */
    int ec;                     /* extended input character */
    int cpos;                   /* current column on screen output */
    char *home_ptr;             /* pointer to home directory string */
    char *ptr;                  /* string pointer */
    char user_name[NSTRING];    /* user name for directory */
    static char buf[NSTRING];   /* buffer to hold tentative name */

    /* if we are executing a command line get the next arg and match it */
    if (clexec) {
        if (get_macro_line(buf) != TRUE)
            return (NULL);
        return (buf);
    }

    /* starting at the beginning of the string buffer */
    cpos = 0;

    /* if it exists, prompt the user for a buffer name */
    if (prompt) {
        if (type == CMP_COMMAND) {
            mlwrite("%s", prompt);
        } else {
            if (defval) {
                mlwrite("%s[%s]: ", prompt, defval);
            } else {
                mlwrite("%s: ", prompt);
            }
        }
    }

    /* build a name string from the keyboard */
    while (TRUE) {

        /* get the keystroke and decode it */
        if (curbp->b_name[0] == '?') {
            if ((curbp->b_name[1] == ' ') &&
                (curbp->b_name[strlen(curbp->b_name) - 1] == '/')) {
                strcpy(buf, curbp->b_name);
                cpos = (int) strlen(curbp->b_name);
                curbp->b_name[0] = ':';
                ec = ' ';
            }
        } else
            ec = mlgetkey();
        c = ectoc(ec);

        /* if it is from a function key, blow it off */
        if (ec & SPEC)
            continue;

        /* if we are at the end, just match it */
        if (c == '\n' || c == '\r') {
            if (defval && cpos == 0)
                return (defval);
            else {
                buf[cpos] = 0;
                return (buf);
            }

        } else if (ec == abortc) {      /* Bell, abort */
            abort_command(FALSE, 0);
            TTflush();
            return (NULL);

        } else if (c == 0x7F || c == 0x08) {    /* rubout/erase */
            if (cpos != 0) {
                mlout('\b');
                mlout(' ');
                mlout('\b');
                --ttcol;
                --cpos;
                TTflush();
            }

        } else if (c == 0x15) { /* C-U, kill */
            while (cpos != 0) {
                mlout('\b');
                mlout(' ');
                mlout('\b');
                --cpos;
                --ttcol;
            }
            TTflush();

        } else if ((c == ' ') || (ec == sterm) || (c == '\t')) {
            /* attempt a completion */
            switch (type) {
            case CMP_BUFFER:
                complete_buffer(buf, &cpos);
                break;
            case CMP_COMMAND:
                complete_command(buf, &cpos);
                break;
            case CMP_FILENAME:
                complete_filename(buf, &cpos);
                break;
            }

            TTflush();
            if (buf[cpos - 1] == 0)
                return (buf);
            goto clist;

        } else if ((cpos > 0) &&
                   ((char) c == DIRSEPCHAR) &&
                   (type == CMP_FILENAME) &&
                   (buf[0] == '~') &&
                   ((home_ptr = getenv("HOME")) != (char *) NULL)) {

            /* save the user name! */
            buf[cpos] = 0;
            strcpy(user_name, &buf[1]);

            /* erase the chars on-screen */
            while (cpos > 0) {
                mlout('\b');
                mlout(' ');
                mlout('\b');
                --cpos;
                --ttcol;
            }

            if (cpos == 0) {

                /* output the home directory */
                ptr = home_ptr;
                while (*ptr) {
                    mlout(*ptr);
                    buf[cpos++] = *ptr++;
                    ++ttcol;
                }

            }

            /* and the last directory seperator */
            if (buf[cpos - 1] != DIRSEPCHAR) {
                mlout(DIRSEPCHAR);
                buf[cpos++] = DIRSEPCHAR;
                ++ttcol;
            }
            TTflush();

        } else if ((cpos > 1) &&
                   ((char) c == DIRSEPCHAR) &&
                   (type == CMP_FILENAME) && (buf[0] == '$')) {

            /* expand an environment variable reference */
            /* save the variable name! */
            buf[cpos] = 0;
            strcpy(user_name, &buf[1]);

            /* erase the chars on-screen */
            while (cpos > 0) {
                mlout('\b');
                mlout(' ');
                mlout('\b');
                --cpos;
                --ttcol;
            }

            ptr = getenv(user_name);
            if (ptr != (char *) NULL) {
                while (*ptr) {
                    mlout(*ptr);
                    buf[cpos++] = *ptr++;
                    ++ttcol;
                }
            }

            /* and the last directory seperator */
            if (buf[cpos - 1] != DIRSEPCHAR) {
                mlout(DIRSEPCHAR);
                buf[cpos++] = DIRSEPCHAR;
                ++ttcol;
            }
            TTflush();

        } else if (c == '?') {

          clist:               /* make a completion list! */
            switch (type) {
            case CMP_BUFFER:
                break;
            case CMP_COMMAND:
                possible_command(buf, &cpos);
                break;
            case CMP_FILENAME:
                break;
            }
            update(TRUE);

            /* if it exists, reprompt the user */
            if (prompt) {
                buf[cpos] = 0;
                if (type == CMP_COMMAND)
                    mlwrite("%s%s", prompt, buf);
                else if (defval)
                    mlwrite("%s[%s]: %s", prompt, defval, buf);
                else
                    mlwrite("%s: %s", prompt, buf);
            }

        } else {
            if (cpos < maxlen && c > ' ') {
                buf[cpos++] = (char) c;
                mlout((int) c);
                ++ttcol;
                TTflush();
            }
        }
    }
}
